"""Simulate fBm trajectories.

This module contains functions used to simulate fractional Brownian motion (fBm) trajectories.
Each dimension is generated with the fBm kernel (Lundahl et al. 1986) using a diffusion and alpha value.
The tracks are made of a mixture of N states following the probabilities to transition.
"""

# Third-party modules
import numpy as np
import pandas as pd
from tqdm import tqdm

# Local functions
from tracksegnet.compute_features import compute_all_features
from tracksegnet.analysis import compute_ptm


def get_fbm(total_frame, alpha, diff):
    """Generates a 1D fBm using the formula from Lundahl et al. 1986.

    :param total_frame: The total number of frames to generate
    :type total_frame: int
    :param alpha: The alpha value for the simulation.
    :type alpha: float
    :param diff: The diffusion value for the simulation.
    :type diff: float
    :return: The 1-dimensional fBm generated track.
    :rtype: np.array
    """
    time_interval = np.arange(0, total_frame, 1)
    x_array, y_array = np.meshgrid(time_interval, time_interval)
    kval = np.abs(x_array - y_array)
    fbm_kernel = diff * (np.abs(kval + 1)**alpha - 2*np.abs(kval)**alpha + np.abs(kval - 1)**alpha)
    fbm = np.linalg.cholesky(fbm_kernel)
    # generate franctional noise from independent Gaussian samples
    fbm = fbm.dot(np.random.normal(0, 1, total_frame))
    return fbm

def generate_fbm_tracks(parms):
    """Creates multiple fBm trajectories with a mixture of states which will be used for training the neural network.

    :param params: Stored parameters containing instructions to generate fBm trajectories.
    :type parms: dict
    :return: The dataframe storing all generated fBm trajectories with keys: `track_id`, `frame`, `x`, `y` and `state`
    :rtype: pd.DataFrame
    """
    # List of initial states (between 0 and parms['num_states']) generated randomly
    init_state_list = np.random.randint(parms['num_states'], size=parms['num_simulate_tracks'])

    # List of track length:
    if parms['track_length_fixed'] is True:
        if parms['track_length'] < parms['length_threshold']:
            parms['track_length'] = parms['length_threshold']
        track_len_list = np.full((parms['num_simulate_tracks']), parms['track_length'])

    else:
        # Length generated randomly from an exponential distribution
        track_len_list = np.random.exponential(parms['beta'], parms['num_simulate_tracks'])
        track_len_list = track_len_list.astype(int) + parms['length_threshold']

    track_df = None
    track_id = 0
    with tqdm(total=parms['num_simulate_tracks']) as pbar:
        for track_len, state_i in zip(track_len_list, init_state_list):
            # Initialization of the new track:
            track = pd.DataFrame({
                'track_id': track_id,
                'frame': range(track_len),
                'x': 0,
                'y': 0,
                'state': 0,
            })
            total = 0
            # The x, y and state values are generated for the whole track length
            while total < track_len:
                # The amount of steps is generated
                if parms['num_states'] == 1:
                    new_steps = track_len - total
                else:
                    new_steps = int(np.random.geometric(1-parms['ptm'][state_i][state_i]))
                        #+ parms['length_threshold']
                    # Condition to make sure the amount of steps does not exceed the track length:
                    if total + new_steps > track_len:
                        new_steps = track_len - total
                        # if new_steps < parms['length_threshold']:
                            # state_i = track['state'][total-1]

                # The same state is attributed for each step
                track.loc[total:total+new_steps-1, 'state'] = state_i
                # The x and y coordinates are generated based on the diffusion parameters:
                alpha = parms['all_states'][state_i]['alpha']
                diff_unitless = parms['all_states'][state_i]['diff'] / parms['unit_diff']
                track.loc[total:total+new_steps-1, 'x'] = get_fbm(new_steps, alpha, diff_unitless)
                track.loc[total:total+new_steps-1, 'y'] = get_fbm(new_steps, alpha, diff_unitless)
                # Update of the current state and the amount of total steps
                if parms['num_states'] > 1:
                    state_i = np.delete(np.arange(parms['num_states']), state_i)\
                        [np.random.randint(parms['num_states']-1)]
                total += new_steps
            track['x'] = np.cumsum(track['x'])
            track['y'] = np.cumsum(track['y'])
            if track_df is None:
                track_df = track
            else:
                track_df = pd.concat((track_df, track))
            pbar.update(1)
            track_id += 1
            # break
    track_df = track_df.reset_index(drop=True)
    return track_df

def run_track_simulation(parms):
    """Run the simulation by first generating fBm trajectories and then computing all features.

    :param params: Stored parameters containing instructions to generate fBm trajectories.
    :type parms: dict
    :return: The dataframe storing all generated fBm trajectories and computed features. Dataframe also saved as a csv file.
    :rtype: pd.DataFrame
    """
    print(f"\nSimulation of {parms['num_simulate_tracks']:,d} trajectories...")
    sim_df = generate_fbm_tracks(parms)
    print(compute_ptm(sim_df, parms))
    compute_all_features(sim_df)
    sim_df.to_csv(parms['simulated_track_path']/'simulated_tracks_df.csv')
    return sim_df



"""Other functions

This module implements a extra utils functions.
"""

# Third-party modules
import os

def check_parms(parms_df):
    """Check each parameter value and raise an error if uncorrect.

    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :return: Refined parameters with type.
    :rtype: dict
    """
    if not os.path.isdir(parms_df['data_path']):
        raise ValueError('data_path does not exist.')

    if parms_df['track_format'] != 'MDF' and parms_df['track_format'] != 'CSV':
        raise ValueError('track_format should be MDF or CSV.')

    try:
        parms_df['time_frame'] = float(parms_df['time_frame'])
    except Exception as exc:
        raise ValueError('time_frame should be a floating value.') from exc

    try:
        parms_df['pixel_size'] = float(parms_df['pixel_size'])
    except Exception as exc:
        raise ValueError('pixel_size should be a floating value.') from exc

    try:
        parms_df['num_states'] = int(parms_df['num_states'])
    except Exception as exc:
        raise ValueError('num_states should be an integer between 2 and 6.') from exc
    if parms_df['num_states'] < 2 or parms_df['num_states'] > 6:
        raise ValueError('num_states should be an integer between 2 and 6.')

    for state in range(1, parms_df['num_states']+1):
        if not f'state_{state}_diff' in parms_df:
            raise ValueError(f'state_{state}_diff not defined in parms.csv')
        if not f'state_{state}_alpha' in parms_df:
            raise ValueError(f'state_{state}_alpha not defined in parms.csv')
        try:
            parms_df[f'state_{state}_diff'] = float(parms_df[f'state_{state}_diff'])
        except Exception as exc:
            raise ValueError(f'state_{state}_diff should be a floating value [dimensionless].') from exc

        try:
            parms_df[f'state_{state}_alpha'] = float(parms_df[f'state_{state}_alpha'])
        except Exception as exc:
            raise ValueError(f'state_{state}_alpha should be a floating value between 0 and 2.') from exc
        if parms_df[f'state_{state}_alpha'] < 0 or parms_df[f'state_{state}_alpha'] > 2:
            raise ValueError(f'state_{state}_alpha should be a floating value between 0 and 2.')
        total = 0
        for state2 in range(1, parms_df['num_states']+1):
            if not f'pt_{state}_{state2}' in parms_df:
                raise ValueError(f'pt_{state}_{state2} not defined in parms.csv')
            try:
                parms_df[f'pt_{state}_{state2}'] = float(parms_df[f'pt_{state}_{state2}'])
            except Exception as exc:
                raise ValueError(f'pt_{state}_{state2} should be a floating value.') from exc
            total += parms_df[f'pt_{state}_{state2}']
        if total != 1:
            raise ValueError(f'Total probability in state {state} should equal 1.')
    return parms_df

def get_color_list(num_states):
    """Create list of colors.
    
    :param num_states: Number of states.
    :type num_states: int
    :return: List of colors for the plots.
    :rtype: list
    """
    if num_states == 1:
        print('Please, select 2 to 6 states.')
    if num_states == 2:
        colors = ['darkblue', 'red']
    elif num_states == 3:
        colors = ['darkblue', 'darkorange', 'red']
    elif num_states == 4:
        colors = ['darkblue', 'darkorange', 'red', 'darkviolet']
    elif num_states == 5:
        colors = ['darkblue', 'darkorange', 'red', 'darkviolet', 'green']
    elif num_states == 6:
        colors = ['darkblue', 'darkorange', 'red', 'darkviolet', 'green', 'k']
    return colors



"""Extract experimental trajectories.

This module contains functions used for the extraction of real experimental trajectories (stored as
MDF or CSV file format).
"""

# Third-party modules
import os
from re import findall, split
from random import uniform
# from multiprocessing import Pool, cpu_count
# from functools import partial
import numpy as np
import pandas as pd
from tqdm import tqdm
from more_itertools import consecutive_groups

# Local functions
from tracksegnet.compute_features import compute_all_features


def fill_gaps_l1(track):
    """Fills the 1-length gaps with an intermediate point having an additional bias (to add randomness).

    :param track: Given track which its 1-length gaps will be filled with an appropriate generated point.
    :type track: pd.DataFrame
    :return: (The updated track with 1-length gaps filled, the number of gaps filled for this given track)
    :rtype: (pd.DataFrame, int)
    """
    # get gaps:
    start = int(track['frame'].iloc[0])
    end = int(track['frame'].iloc[-1])
    missing_frames = sorted(set(range(start, end + 1)).difference(track['frame']))
    gaps = [list(group) for group in consecutive_groups(missing_frames)]
    num_gaps_l1 = 0
    for _, gap in enumerate(gaps):
        if len(gap) != 1:
            continue
        num_gaps_l1 += 1
        missing = gap[0]
        prev_point = track[track['frame'] == missing-1].iloc[0].to_dict()
        next_point = track[track['frame'] == missing+1].iloc[0].to_dict()
        average_points = {'x': np.mean([prev_point['x'], next_point['x']]),
                          'y': np.mean([prev_point['y'], next_point['y']])}
        displ = {'x': abs(prev_point['x'] - next_point['x']),
                 'y': abs(prev_point['y'] - next_point['y'])}
        bias = {'x': uniform(-displ['x']/4, displ['x']/4),
                'y': uniform(-displ['y']/4, displ['y']/4)}
        if displ['x'] == 0 and displ['y'] == 0:
            new_point = pd.DataFrame({
                'index': 'X',
                'x': average_points['x'],
                'y': average_points['y'],
                'frame': np.array([missing]),
                'data folder': track.iloc[0]['data folder'],
                'track_id': track.iloc[0]['track_id']
            })
        else:
            if displ['x'] != 0:
                slop = (average_points['y'] - prev_point['y'])\
                      / (average_points['x'] - next_point['x'])
            else:
                slop = (average_points['y'] - average_points['y'])\
                      / (average_points['x'] - average_points['x']+1)
            new_point = pd.DataFrame({
                'index': 'X',
                'x': average_points['x'] - bias['x'],
                'y': average_points['y'] - bias['x']*slop - bias['y'],
                'frame': np.array([missing]),
                'data folder': track.iloc[0]['data folder'],
                'track_id': track.iloc[0]['track_id']
            })
        track = pd.concat(
            (track[track['frame'] < missing], new_point, track[track['frame'] > missing])
        )
    return track, num_gaps_l1

def fill_gaps(parms, track_df):
    """Fill the gap in trajectory frames by taking an intermediate point and splits the trajectory into two separate tracks for gaps larger than 1 frame.

    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param track_df: Dataframe containing all extracted trajectories with keys: `x`, `y`, `frame`, `data_folder`, `track_id`.
    :type track_df: pd.DataFrame
    :return: The updated dataframe with filled gaps or splitted tracks.
    :rtype: pd.DataFrame
    """
    num_tracks = len(track_df['track_id'].unique())
    print(f'\nCheck and fill gaps in {num_tracks:,d} trajectories...')
    track_df_2 = None
    total_gaps_l1, total_gaps_ln = 0, 0
    for track_id in tqdm(track_df['track_id'].unique()):
        track = track_df[track_df['track_id'] == track_id]
        # fill gaps with one point coord missing:
        new_track, num_gaps_l1 = fill_gaps_l1(track)
        new_track = new_track.reset_index(drop=True)
        total_gaps_l1 += num_gaps_l1
        # The Track is splitted for gaps of length-2 or more:
        sta = new_track.iloc[0].name
        end = new_track.iloc[-1].name
        indexes = new_track[new_track['frame'].diff(periods=-1) < -1].index.to_list()
        total_gaps_ln += len(indexes)
        indexes.append(end)
        for i, ind in enumerate(indexes):
            # new 'track_id':
            new_track.loc[sta:ind, 'track_id'] = split('_0$', new_track.iloc[0]['track_id'])[0]+f'_{str(i)}'
            sta = ind+1
        if track_df_2 is None:
            track_df_2 = new_track
        else:
            track_df_2 = pd.concat((track_df_2, new_track))
    # Keep only tracks longer than parms['length_threshold']:
    print(f'\nKeep trajectories longer than {parms["length_threshold"]}..')
    track_df_3 = None
    for track_id in tqdm(track_df_2['track_id'].unique()):
        track = track_df_2[track_df_2['track_id'] == track_id]
        if len(track) >= parms['length_threshold']:
            if track_df_3 is None:
                track_df_3 = track
            else:
                track_df_3 = pd.concat((track_df_3, track))
        else:
            total_gaps_ln -= 1
    print(f"{total_gaps_l1:,d} length-1 gaps were filled in", end=' ')
    print(f"and {total_gaps_ln:,d} tracks created due to length-2+ gaps.")
    return track_df_3.reset_index(drop=True)

def extract_1_mdf(folder_name, parms, filename='tracks.simple.mdf'):
    """Extracts experimental trajectories from a single MDF file and returns a pandas DataFrame which stores
    the trajectories.

    :param folder_name: Name of the folder containing the trajectory file.
    :type folder_name: str
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param filename: File which stores trajectory coordinates extracted from the tracking.
                     `[Defaults: 'tracks.simple.mdf']`
    :type filename: str
    :return: All extracted trajectories as a dataframe with keys: `x`, `y`, `frame`, `data_folder`, `track_id`.
    :rtype: pd.DataFrame
    """    
    if not os.path.isfile(folder_name/filename):
        for filename in os.listdir(folder_name):
            if filename.endswith(".mdf"):
                break
    # Initialization
    index = 0
    track_df = None
    # coordinates of the current track:
    current_coord = {'frame': np.array([]), 'x': np.array([]), 'y': np.array([])}
    with open(folder_name/filename, 'r', encoding='utf-8') as file:
        for line in file.readlines():
            # Find a new Track
            if line[:5] == 'Track':
                # If old Track longer than the threshold then it is added to the list
                if len(current_coord['x']) >= parms['length_threshold']:
                    track = pd.DataFrame(current_coord)
                    track['data folder'] = str(folder_name) # PosixPath to str
                    track['track_id'] = folder_name.name.split('_')[-1] + '_' + index
                    if track_df is None:
                        track_df = track
                    else:
                        track_df = pd.concat((track_df, track))
                # Initialize new Track
                index = findall(r'\d+', line)[0]
                current_coord = {'x': np.array([]), 'y': np.array([]), 'frame': np.array([])}
            # Add Point to the current Track
            elif line[:5] == 'Point':
                contents = line.split(' ')
                current_coord['x'] = np.append(current_coord['x'], float(contents[2]))
                current_coord['y'] = np.append(current_coord['y'], float(contents[3]))
                current_coord['frame'] = np.append(current_coord['frame'], int(contents[5]))
            # End of the file, the last Track is added
            elif line[:3] == 'End':
                # If old Track longer than the threshold then it is added to the list
                if len(current_coord['x']) >= parms['length_threshold']:
                    track = pd.DataFrame(current_coord)
                    track['data folder'] = str(folder_name) # PosixPath to str
                    track['track_id'] = folder_name.name.split('_')[-1] + '_' + index
                    if track_df is None:
                        track_df = track
                    else:
                        track_df = pd.concat((track_df, track))
    return track_df

def extract_1_csv(folder_name, parms, filename='tracks.csv'):
    """Extracts experimental trajectories from a single CSV file and returns a pandas DataFrame which stores
    the trajectories.

    :param folder_name: Name of the folder containing the trajectory file.
    :type folder_name: str
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param filename: File which stores trajectory coordinates extracted from the tracking.
                     `[Defaults: 'tracks.csv']`
    :type filename: str
    :return: All extracted trajectories as a dataframe with keys: `x`, `y`, `frame`, `data_folder`, `track_id`.
    :rtype: pd.DataFrame
    """ 
    if not os.path.isfile(folder_name/filename):
        for filename in os.listdir(folder_name):
            if filename.endswith(".csv"):
                break
    tracks = pd.read_csv(filename)
    tracks['data folder'] = str(folder_name)

    track_df = None
    for track_id in tracks['track_id'].unique():
        track = tracks[tracks['track_id'] == track_id]
        if len(track['x']) >= parms['length_threshold']:
            track['track_id'] = folder_name.name.split('_')[-1] + '_' + track_id
            if track_df is None:
                track_df = track
            else:
                track_df = pd.concat((track_df, track))
    return track_df

def extract_all_tracks(parms):
    """Extracts the trajectories from several MDF (or CSV) files, fills the gaps and return a unique pandas
    DataFrame containing tracks.

    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :return: All extracted trajectories as a dataframe with keys: `x`, `y`, `frame`, `data_folder`, `track_id`.
    :rtype: pd.DataFrame
    """
    num_files = len(parms['folder_names'])
    print(f"\nExtraction of tracjectories from {num_files} mdf files...")
    all_track_df = []
    if parms['track_format'] == 'MDF':
        for folder_name in tqdm(parms['folder_names']):
            all_track_df.append(extract_1_mdf(folder_name, parms))
    elif parms['track_format'] == 'CSV':
        for folder_name in tqdm(parms['folder_names']):
            all_track_df.append(extract_1_csv(folder_name, parms))
    track_df = pd.concat(all_track_df).reset_index()
    print(f'\n{track_df.shape[0]} coordinate points')
    print(track_df.head())
    track_df_2 = fill_gaps(parms, track_df)
    print(f'\n{track_df_2.shape[0]} coordinate points')
    print(track_df_2.head())
    compute_all_features(track_df_2)
    print(track_df_2.head())
    track_df_2.to_csv(parms['result_path']/f"{parms['data_path'].name}_tracks_df.csv", index=False)
    return track_df_2

def predict_states(track_df, model, parms):
    """Predicts the states for each trajectory using a trained model and saves the resulting state prediction as a csv file.

    :param track_df: Name of the folder containing the trajectory file.
    :type track_df: pd.DataFrame
    :param model: Trained neural network.
    :type model: keras model
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    """
    num_tracks = len(track_df['track_id'].unique())
    print(f'\nState prediction of {num_tracks:,d} trajectories...')
    for track_id in tqdm(track_df['track_id'].unique()):
        track = track_df[track_df['track_id'] == track_id]
        features = track[['displ_x', 'displ_y', 'dist', 'mean_dist_1', 'mean_dist_2', 'angle']]
        features = features.to_numpy()[1:-1]
        features = features.reshape(1, features.shape[0], features.shape[-1])
        predicted_states = model.predict(features, verbose=0)
        predicted_states = [np.nan] + predicted_states.argmax(axis=2)[0].tolist() + [np.nan]
        track_df.loc[track.index, 'state'] = predicted_states
    print('')
    print(track_df.head())
    track_df.to_csv(parms['result_path']/f"{parms['data_path'].name}_tracks_df.csv", index=False)



"""Analysis and plotting functions

This modules contains functions for trajectory analysis and plotting.
"""

# Third-party modules
import warnings
import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.optimize import curve_fit
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from matplotlib import cm
import seaborn as sns

matplotlib.rcParams.update({
    'font.size': 12,
    'axes.labelsize': 13,
    'figure.labelsize': 13,
    'mathtext.fontset': 'dejavusans'
})

def convert_sigma(diffusion, parms):
    """Converts diffusion to sigma value.

    :param diffusion: Diffusion value [um**2/s] to convert.
    :type diffusion: float
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :return: sigma value.
    :rtype: float
    """
    sigma = np.sqrt((diffusion/parms['unit_diff'])*2)
    return sigma

def convert_diffusion(sigma, parms):
    """Converts sigma to diffusion value.

    :param sigma: Sigma value to convert.
    :type sigma: float
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :return: Diffusion value [um**2/s].
    :rtype: float
    """
    diffusion = ((sigma**2)/2)*parms['unit_diff']
    return diffusion

def compute_ptm(track_df, parms):
    """Computes the probability transition matrix (PTM).

    :param track_df: Dataframe containing all extracted trajectories.
    :type track_df: pd.DataFrame
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :return: probabilities to transition between states
    :rtype: pd.DataFrame
    """
    print('\nCompute probability transition matrix (PTM)...')
    counts = np.zeros((parms['num_states'], parms['num_states']), dtype=int)
    for track_id in tqdm(track_df['track_id'].unique()):
        track = track_df[track_df['track_id'] == track_id]
        for i, j in zip(track['state'], track['state'][1:]):
            if pd.notna(i) and pd.notna(j):
                # print(i, j)
                counts[int(i)][int(j)] += 1
    # estimate Probs
    sum_of_rows = counts.sum(axis=1)
    # if some row-sums are zero, replace with any number becasue the sum in in denomimator and
    # the numerator will be 0 in any case, so we will get the correct 0 for p_ij in the result
    sum_of_rows[sum_of_rows == 0] = 1
    pis = counts / sum_of_rows[:, None]
    ptm = pd.DataFrame(pis)
    ptm.columns = ptm.columns+1
    ptm['num'] = sum_of_rows
    ptm.to_csv(parms['result_path']/'transition_probabilities.csv', index=False)
    return ptm

def make_tracklet_lists(track_df, parms):
    """Segments the trajectories into tracklet (based on the state group) and regroup them within the
    same list.

    :param track_df: Dataframe containing all extracted trajectories.
    :type track_df: pd.DataFrame
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :return: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :rtype: list
    """
    print('\nSegment and group tracklets per diffusive states...')
    tracklet_lists = [[] for _ in range(parms['num_states'])]
    for track_id in tqdm(track_df['track_id'].unique()):
        track = track_df[track_df['track_id'] == track_id]
        track = track.dropna()
        ind = np.where(np.diff(track['state']) != 0)[0]
        ind = np.insert(ind, len(ind), len(track)-1)
        tmp = 0
        for i in ind:
            tracklet_tmp = track[tmp:i+1]
            state = tracklet_tmp.iloc[0]['state']
            if not pd.notna(state):
                print(tracklet_tmp)
                continue
            tracklet_lists[int(state)].append(tracklet_tmp)
            tmp = i+1
    # Check states
    for state in range(parms['num_states']):
        for track in tracklet_lists[state]:
            mean_states = np.mean(track['state'])
            if mean_states not in [float(i) for i in range(parms['num_states'])]:
                print(mean_states)
    return tracklet_lists

def compute_msd(track, size, dim=2):
    """Computes the mean square displacement (MSD) for a given track in order to estimate D and alpha
    using the formula:  log(MSD(dt)) ~ alpha.log(dt) + log(C), with C = 2nD.
    
    :param track: Dataframe containing a trajectory's coordinates.
    :type track: pd.DataFrame
    :param size: Number of delta time points to use for the MSD curve fit.
    :type size: int
    :param dim: Dimentionality of the track [Defaults: 2].
    :type dim: int
    :return: (msd values, time axis, measured alpha (slope), measured log_C (intercept), resulting diffusion)
    :rtype: (np.array, np.array, float, float, float)
    """
    def f_slope_intercept(x_val, a_val, b_val):
        """Linear regression y = ax + b."""
        return a_val*x_val + b_val
    if size <= 2:
        size = 3
    coords = {'x': track['x'].to_numpy(), 'y': track['y'].to_numpy()}
    delta_array = np.arange(1, size+1)
    msd = np.zeros(delta_array.size)
    sigma = np.zeros(delta_array.size)
    for i, delta in enumerate(delta_array):
        if dim == 2:
            x_displ = coords['x'][delta:]-coords['x'][:-delta]
            y_displ = coords['y'][delta:]-coords['y'][:-delta]
            res = abs(x_displ)**2 + abs(y_displ)**2
            msd[i] = np.mean(res)
            sigma[i] = np.std(res)
        else:
            print('Dimension should be 2.')
    # popt, _ = curve_fit(f_slope_intercept, np.log(delta_array), np.log(msd))
    popt, _ = curve_fit(f_slope_intercept, np.log(delta_array), np.log(msd), sigma=sigma,
                        absolute_sigma=True)
    alpha = popt[0] # slope
    log_c = popt[1] # intercept
    diffusion = np.exp(log_c)/(2*dim)
    return (msd, delta_array, alpha, log_c, diffusion)

def compute_all_msd(tracklet_lists, parms):
    """Computes the MSD for each given track.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :return: Dictionary containing lists of measured motion parameters for each track with keys: alpha, diffusion and track_id.
    :rtype: dict
    """
    print('\nCompute MSD...')
    motion_parms = [{'alpha': [], 'diffusion': [], 'track_id': []}\
                    for _ in range(parms['num_states'])]
    for state in tqdm(range(parms['num_states'])):
        for tracklet in tracklet_lists[state]:
            if len(tracklet) >= parms['length_threshold']:
                _, _, alpha, _, diffusion = compute_msd(tracklet, size=4)
                diffusion = diffusion * parms['unit_diff']
                # print(f"spot {tracklet['track_id'].iloc[0]},", end=' ')
                # print(f"L = {len(tracklet)}, alpha = {alpha:.3}, diffusion = {diffusion:.3}")
                motion_parms[state]['alpha'].append(alpha)
                motion_parms[state]['diffusion'].append(diffusion)
                motion_parms[state]['track_id'].append(tracklet['track_id'].iloc[0])
        pd.DataFrame(motion_parms[state]).to_csv(
            parms['result_path']/f"motion_parms_state_{state+1}.csv", index=False)
    return motion_parms

def plot_scatter_alpha_diffusion(motion_parms, parms):
    """Makes a scatterplot of the alpha and diffusion distributions.

    :param motion_parms: Dictionary containing lists of measured motion parameters for each track with keys: alpha, diffusion and track_id.
    :type motion_parms: dict
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    """
    print('\nPlot D/alpha scatterplot...')
    print('centroids...')
    func = None
    for state in range(parms['num_states']):
        col = parms['colors'][state]
        dataframe = pd.DataFrame({
            'x': motion_parms[state]['diffusion'],
            'y': motion_parms[state]['alpha']
        })
        if func is None:
            func = sns.JointGrid(x='x', y='y', data=dataframe, height=6.5)
            axs = func.ax_joint
        tmp_diff = np.array(motion_parms[state]['diffusion'])
        tmp_alpha = np.array(motion_parms[state]['alpha'])
        tmp_diff = tmp_diff[tmp_alpha >= 0]
        tmp_alpha = tmp_alpha[tmp_alpha >= 0]
        axs.scatter(tmp_diff, tmp_alpha, alpha=0.6, color=col, edgecolor='none', s=2)
        centroid = {'diffusion': np.median(tmp_diff), 'alpha': np.median(tmp_alpha)}
        print(f"state {state+1}: alpha = {centroid['alpha']:>4.2f},", end=" ")
        print(f"D = {centroid['diffusion']:>4.2f} µm²/s")
        axs.plot(centroid['diffusion'], centroid['alpha'], 'o', color='0.1', alpha=1, ms=3)
        axs.plot(centroid['diffusion'], centroid['alpha'], 'o', color='0.1', alpha=1, ms=1, mew=8)
        # bins=150, bins=200)
        bins = np.logspace(-2.5, 1.5, 200)
        sns.histplot(x=tmp_diff, color=col, ax=func.ax_marg_x, alpha=0.5, bins=bins, stat='count')
        bins = np.linspace(0, 2, 100)
        sns.histplot(y=tmp_alpha, color=col, ax=func.ax_marg_y, alpha=0.5, bins=bins, stat='count')
    axs.set_xscale('log')
    axs.set_xlabel(r'$\mathrm{D}$ $[\mu m^2/s]$')
    axs.set_ylabel(r'$\mathrm{\alpha}$')
    func.ax_marg_x.grid(axis='x', alpha=0.5)
    func.ax_marg_y.grid(axis='y', alpha=0.5)
    axs.set_xlim([10**-2, 10**0.5])
    axs.set_ylim([0, 2])
    axs.grid(alpha=0.5)
    fig = plt.gcf()
    # fig.suptitle(f"{parms['data_path'].name} dataset")
    fig.tight_layout()
    fig.subplots_adjust(wspace=0, hspace=0)
    plt.savefig(f"{parms['result_path']}/fig_{parms['data_path'].name}_scatterplot_msd"+\
                f".{parms['fig_format']}",
                dpi=parms['fig_dpi'], transparent=parms['fig_transparent'])
    plt.close()

def plot_proportion(tracklet_lists, parms):
    """Plots the proportion of tracklets in each diffusive state as a pie chart.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    """
    print('\nPlot proportion...')
    tracklet_num = [len(tracklet_list) for tracklet_list in tracklet_lists]
    print(tracklet_num)
    labels = [f'state {state+1}\nn = {len(tracklet_list)}'\
              for state, tracklet_list in enumerate(tracklet_lists)]
    fig, axs = plt.subplots(1, figsize=(4, 4))
    axs.pie(tracklet_num, labels=labels, autopct='%1.1f%%', colors=parms['colors'], wedgeprops={"alpha": 0.6})
    axs.axis('equal')
    fig.suptitle(f"{parms['data_path'].name} dataset")
    fig.tight_layout()
    plt.savefig(f"{parms['result_path']}/fig_{parms['data_path'].name}_tracklet_proportion."+\
                f"{parms['fig_format']}",
                dpi=parms['fig_dpi'], transparent=parms['fig_transparent'])
    plt.close()

def compute_displ(tracklet_lists, parms, dtime=1):
    """Computes the displacements for each tracklet state.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param dtime: Time interval between two data points [Default: 1]
    :type dtime: int
    :return: List of calculated displacements for each tracklet group.
    :rtype: list
    """
    all_displ_tracklet = [{dtime: None} for _ in range(len(tracklet_lists))]
    for state in tqdm(range(parms['num_states'])):
        for tracklet in tracklet_lists[state]:
            displ_x = tracklet[dtime:].reset_index(drop=True)['x']-\
                tracklet.reset_index(drop=True)['x']
            displ_x = displ_x.to_numpy().astype(np.float64)
            displ_y = tracklet[dtime:].reset_index(drop=True)['y']-\
                tracklet.reset_index(drop=True)['y']
            displ_y = displ_y.to_numpy().astype(np.float64)
            if all_displ_tracklet[state][dtime] is None:
                all_displ_tracklet[state][dtime] = {'x': displ_x, 'y': displ_y}
            else:
                all_displ_tracklet[state][dtime]['x'] = np.append(
                    all_displ_tracklet[state][dtime]['x'], displ_x)
                all_displ_tracklet[state][dtime]['y'] = np.append(
                    all_displ_tracklet[state][dtime]['y'], displ_y)

        all_displ_tracklet[state][dtime]['x'] = all_displ_tracklet[state][dtime]['x']\
            [~np.isnan(all_displ_tracklet[state][dtime]['x'])]*parms['pixel_size']
        all_displ_tracklet[state][dtime]['y'] = all_displ_tracklet[state]\
            [dtime]['y'][~np.isnan(all_displ_tracklet[state][dtime]['y'])]*parms['pixel_size']
    return all_displ_tracklet

def plot_displ(tracklet_lists, parms, dtime=1):
    """Plots the distribution of displacements per tracklet state.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param dtime: Time interval between two data points [Default: 1]
    :type dtime: int
    """
    print(f'\nPlot displacements with dt={dtime}...')
    all_displ_tracklet = compute_displ(tracklet_lists, parms, dtime=dtime)

    plt.close()
    hist = {}
    for state, all_displ_tracklet_state in enumerate(all_displ_tracklet):
        displ_tmp = np.concatenate((all_displ_tracklet_state[dtime]['x'],
                                    all_displ_tracklet_state[dtime]['y']))
        binwidth = 0.02
        bins = np.arange(-0.4-0.01, 0.4 + binwidth, binwidth)
        sns.histplot(displ_tmp, bins=bins, stat='density')
        bars = plt.gca().patches[0]
        xy_coords = np.array([[bars.get_x()+(bars.get_width()/2), bars.get_height()]\
                      for bars in plt.gca().patches]).T
        hist[state] = xy_coords
        plt.close()

    fig, axs = plt.subplots(1, figsize=(5, 5))
    axs.grid(alpha=0.5)

    for state, _ in enumerate(all_displ_tracklet):
        axs.plot(hist[state][0], hist[state][1], 'o-', color=parms['colors'][state],
                 label=f'state {state+1}', ms=5.5, lw=1)
    axs.set_ylabel('Probability Density')
    axs.set_xlabel(r'$\mathrm{\Delta r}$ $[\mu m]$')
    axs.set_title('Distribution of X and Y displacements')
    axs.legend(loc='upper left')
    fig.tight_layout()
    plt.savefig(f"{parms['result_path']}/fig_{parms['data_path'].name}_displ"+\
                f".{parms['fig_format']}",
                dpi=parms['fig_dpi'], transparent=parms['fig_transparent'])
    plt.close()

def compute_angles(tracklet_lists, parms, dtime=1):
    """Computes the angles for each tracklet state.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param dtime: Time interval between two data points [Default: 1]
    :type dtime: int
    :return: Dictionary of numpy arrays for each tracklet group containing the calculated angles.
    :rtype: dict
    """
    tracklet_angles = {}
    for state in tqdm(range(parms['num_states'])):
        angles_current_state = np.array([])
        for tracklet in tracklet_lists[state]:
            displ_x = tracklet[dtime:].reset_index()['x']-tracklet.reset_index()['x']
            displ_x = displ_x.to_numpy().astype(np.float64)
            displ_y = tracklet[dtime:].reset_index()['y']-tracklet.reset_index()['y']
            displ_y = displ_y.to_numpy().astype(np.float64)
            # calulate angles
            alpha_1 = np.arctan2(displ_y[:-1], displ_x[:-1])
            alpha_2 = np.arctan2(displ_y[1:], displ_x[1:]) + 2*np.pi
            angles = (alpha_2 - alpha_1) % (2*np.pi)
            angles[angles > np.pi] = angles[angles > np.pi] - (2*np.pi)
            angles = angles[~np.isnan(angles)]
            angles_current_state = np.append(angles_current_state, angles)
        tracklet_angles[state] = angles_current_state
    return tracklet_angles

def calculate_fold_180_0(angle_list):
    """Calculates the fold anisotropy for a given list of angles. ref doi:10.1038/s41589-019-0422-3

    :param angle_list: Numpy array containing the calculated angles for a given tracklet group.
    :type angle_list: dict
    :return: (measure fold anisopropy, total number of 180 +/-30° angles, total number of 0 +/-30° angles)
    :rtype:  (float, int, int)
    """
    num = {'180': 0, '0': 0}
    for angle in angle_list:
        # angle = angle + 2*np.pi
        if ((5*np.pi)/6 <= angle <= np.pi) or (-np.pi <= angle <= (-5*np.pi)/6):
            num['180'] += 1
        elif (0 <= angle <= np.pi/6) or (-np.pi/6 <= angle <= 0):
            num['0'] += 1
    try:
        fold_180_0 = num['180'] / num['0']
    except ZeroDivisionError:
        fold_180_0 = None
    return (fold_180_0, (num['180'], num['0']))

def plot_angles(tracklet_lists, parms, dtime=1):
    """Plots the angular distribution per tracklet state.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param dtime: Time interval between two data points [Default: 1]
    :type dtime: int
    """
    print(f'\nPlot angles with dt={dtime}...')
    tracklet_angles = compute_angles(tracklet_lists, parms, dtime=dtime)

    plt.close()
    hist = {}
    for state, _ in enumerate(tracklet_angles):
        sns.histplot(tracklet_angles[state], bins=25, stat='density',
                     color=parms['colors'][state], alpha=0.5)
        bars = plt.gca().patches[0]
        xy_coords = np.array([[bars.get_x()+(bars.get_width()/2), bars.get_height()]\
                      for bars in plt.gca().patches]).T
        hist[state] = xy_coords
        plt.close()

    fig, axs = plt.subplots(1, figsize=(5, 5))
    axs.grid(alpha=0.5)

    formula = r'$\mathrm{f_{180/0}}$'
    labels = []
    for state, _ in enumerate(tracklet_lists):
        axs.plot(hist[state][0], hist[state][1], 'o-', color=parms['colors'][state],
                 label=f'state {state+1}',
                 ms=5.5, lw=1)
        fold = calculate_fold_180_0(tracklet_angles[state])[0]
        labels.append(mlines.Line2D([], [], color=parms['colors'][state],
                                    label=f"{formula} = {fold:.2}", lw=1))
    axs.set_xlim(-np.pi, np.pi)
    # axs.set_ylim(0.05, 0.30)
    axs.set_xticks(np.arange(-np.pi, np.pi+0.01, np.pi/2))
    axs.set_xticklabels([r'$\mathrm{-180^o}$', r'$\mathrm{-90^o}$', r'$\mathrm{0^o}$',
                         r'$\mathrm{90^o}$', r'$\mathrm{180^o}$'])
    axs.legend(handles=labels, loc='upper center')
    axs.set_xlabel(r'$\mathrm{\theta}$')
    axs.set_ylabel('Probability Density')
    axs.set_title('Angular distributions')
    fig.tight_layout()
    plt.savefig(f"{parms['result_path']}/fig_{parms['data_path'].name}_angles"+\
                f".{parms['fig_format']}",
                dpi=parms['fig_dpi'], transparent=parms['fig_transparent'])
    plt.close()

def compute_vac(tracklet_lists, parms, dtime=1, thr=1000):
    """Computes the velocity autocorrelation (VAC) curves for each tracklet state.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param dtime: Time interval between two data points [Default: 1]
    :type dtime: int
    :param thr: Threshold for the vac limiting the number of points calculated [Default: 1000].
    :type thr: int
    :return:  List of calculated VAC for each tracklet group.
    :rtype: list
    """
    time_frame_ms = parms['time_frame']*1000
    all_vac = [{dtime: None} for _ in range(len(tracklet_lists))]
    for state in tqdm(range(parms['num_states'])):
        vac_tmp = []
        for tracklet in tracklet_lists[state]:
            if len(tracklet) <= dtime:
                continue
            xcoord = tracklet['x'].to_numpy()
            vac_tmp.append(np.dot(((xcoord[dtime:]-xcoord[:-dtime])/dtime)*time_frame_ms,
                                  ((xcoord[dtime]-xcoord[0])/dtime)*time_frame_ms))
            ycoord = tracklet['y'].to_numpy()
            vac_tmp.append(np.dot(((ycoord[dtime:]-ycoord[:-dtime])/dtime)*time_frame_ms,
                                  ((ycoord[dtime]-ycoord[0])/dtime)*time_frame_ms))
        vac = np.zeros((len(vac_tmp), thr))
        vac[:] = np.nan
        for i, _ in enumerate(vac):
            vac[i][:len(vac_tmp[i])] = vac_tmp[i][:1000]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            all_vac[state][dtime] = np.nanmean(vac, axis=0)
    return all_vac

def plot_vac(tracklet_lists, parms, dtime=1):
    """Plots the velocity autocorrelation curves per tracklet state.

    :param tracklet_lists: List of tracklet groups. Each group consists of a dataframe of tracklets.
    :type tracklet_lists: list
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    :param dtime: Time interval between two data points [Default: 1]
    :type dtime: int
    """
    print(f'\nPlot velocity autocorrelation with dt={dtime}...')
    all_vac = compute_vac(tracklet_lists, parms, dtime=dtime)

    fig, axs = plt.subplots(1, figsize=(5, 5))
    axs.grid(alpha=0.5)
    # plot theoretical curves
    xaxis = np.arange(0, 100)
    alphas = np.arange(0.1, 1.3, 0.1)
    cmap = cm.get_cmap("viridis_r")
    cmap = cmap(np.linspace(0, 1, len(alphas)))
    thr_prev = None
    for i, alpha_thr in enumerate(alphas[::-1]):
        thr = ((xaxis+1)**(alpha_thr) + abs(xaxis-1)**(alpha_thr) - 2*(xaxis**(alpha_thr)))/2
        if thr_prev is None:
            thr_prev = thr.copy()
        else:
            axs.fill_between(xaxis*parms['time_frame']*1000, thr, thr_prev, facecolor=cmap[i],
                             alpha=1)
            thr_prev = thr.copy()

    for state, _ in enumerate(tracklet_lists):
        x_axis_norm = np.array([i*parms['time_frame']*1000\
                               for i in range(len(all_vac[state][dtime]))])
        axs.plot(x_axis_norm, (all_vac[state][dtime]/all_vac[state][dtime][0]), 'o-',
                 color=parms['colors'][state], lw=1, ms=5.5, label=f'state {state}')
    axs.set_ylabel(r'$\mathrm{C^{(\delta=1)}_{v}(\Delta t)}$'+\
                   r'$\mathrm{/}$ $\mathrm{C^{(\delta=1)}_{v}(0)}$')
    axs.set_xlabel(r'$\mathrm{\Delta t}$')#' $[ms]$')
    axs.set_xlim([0, 30])
    axs.set_xticks(np.arange(0, parms['time_frame']*1000*5, parms['time_frame']*1000))
    axs.set_ylim([-0.5, 1])
    # axs.legend(loc='upper right', fontsize='x-small')
    thr_curves = plt.contourf([[0, 0], [0, 0]], alphas, cmap='viridis')
    fig.colorbar(thr_curves, ax=axs, ticks=alphas).\
                 set_label(label=r'$\mathrm{\alpha}$ value of theoretical fBm')
    axs.set_title('Velocity autocorrelation curves')
    fig.tight_layout()
    plt.savefig(f"{parms['result_path']}/fig_{parms['data_path'].name}_vac"+\
                f".{parms['fig_format']}",
                dpi=parms['fig_dpi'], transparent=parms['fig_transparent'])
    plt.close()



"""Compute track features.

This module contains functions computing different features which will be used as input of the deep-learning network (for both training and testing).
"""

# Third-party modules
import numpy as np
from tqdm import tqdm


def compute_displacements(track, delta):
    """Computes x and y displacements between each pair of points i and i+delta.

    :param track: Given track where displacements will be computed along its entire length.
    :type track: pd.DataFrame
    :param delta: Lag interval between frames.
    :type delta: int
    :return: Computed (x displacements, y displacements) along the track.
    :rtype: (np.array, np.array)
    """
    displ_x = track[delta:].reset_index(drop=True)['x']-track.reset_index(drop=True)['x']
    displ_y = track[delta:].reset_index(drop=True)['y']-track.reset_index(drop=True)['y']
    return displ_x.to_numpy(), displ_y.to_numpy()

def compute_dist(displ_x, displ_y):
    """Computes distances based on x and y displacements.

    :param displ_x: x displacements of a given track.
    :type displ_x: np.array
    :param displ_y: y displacements of a given track.
    :type displ_y: np.array
    :return: Computed distances along the track.
    :rtype: np.array
    """
    return np.sqrt(displ_x**2 + displ_y**2)

def compute_mean_distances(track, delta, num=1):
    """Computes a mean of distances from point i-num to point i+num.

    :param track: Given track where mean of distances will be computed along its entire length.
    :type track: pd.DataFrame
    :param delta: Lag interval between frames while measuring the displacements.
    :type delta: int
    :param num: Number of neighboring point for the average. [Defaults: +/- 1 close neighbors]
    :type num: int
    :return: The mean of distances.
    :rtype: np.array
    """
    displ_x, displ_y = compute_displacements(track, delta)
    dist = compute_dist(displ_x, displ_y)
    num_frames = len(displ_x)
    mean_dist = []
    for i in range(num_frames):
        start = i - num
        start = max(start, 0)
        end = i + num
        if end > num_frames - num - delta:
            end = num_frames - num - delta
        # print(dist[start:end+1], end='\n\n')
        if len(dist[start:end+1]) < 1:
            mean_dist.append(np.nan)
        else:
            mean_dist.append(np.mean(dist[start:end+1]))
    return mean_dist

def compute_angles(displ_x, displ_y):
    """Computes angles using two consecutive displacements using x and y 1D-displacements.

    :param displ_x: x displacements of a given track.
    :type displ_x: np.array
    :param displ_y: y displacements of a given track.
    :type displ_y: np.array
    :return: Angles computed along the track.
    :rtype: np.array
    """
    alpha_1 = np.arctan2(displ_y[:-1], displ_x[:-1])
    alpha_2 = np.arctan2(displ_y[1:], displ_x[1:]) + 2*np.pi
    angles = (alpha_2 - alpha_1) % (2*np.pi)
    angles[angles > np.pi] = angles[angles > np.pi] - (2*np.pi)
    angles = np.insert(angles, 0, np.nan)
    return angles

def compute_all_features(track_df):
    """Computes all features necessary for the inputs of the deep-learning network.
    The features are directly added to the dataframe of trajectories.

    :param track_df: All trajectories as a dataframe with keys: `x`, `y`, `frame`, `data_folder`, `track_id`.
    :type track_df: pd.DataFrame
    """
    print('\nCompute features...')
    track_df[['displ_x', 'displ_y', 'dist', 'mean_dist_1', 'mean_dist_2', 'angle']] = np.nan
    for track_id in tqdm(track_df['track_id'].unique()):
        track = track_df[track_df['track_id'] == track_id]
        displ_x, displ_y = compute_displacements(track, delta=1)
        track_df.loc[track.index, 'displ_x'] = displ_x
        track_df.loc[track.index, 'displ_y'] = displ_y
        track_df.loc[track.index, 'dist'] = compute_dist(displ_x, displ_y)
        track_df.loc[track.index, 'mean_dist_1'] = compute_mean_distances(track, delta=1)
        track_df.loc[track.index, 'mean_dist_2'] = compute_mean_distances(track, delta=2)
        track_df.loc[track.index, 'angle'] = compute_angles(displ_x, displ_y)



#!/usr/bin/env python
# coding: utf-8

"""
    Segmentation of trajectories into N diffusive states using LSTM neural network classifier

    Usage:
        # Adjust parameters in parms.csv before running the following command on the terminal:
        python tracksegnet-main.py parms.csv
"""

# Third-party modules
import os
import sys
from pathlib import Path
from datetime import datetime
import numpy as np
import pandas as pd
from tensorflow.keras.models import load_model

# Local modules
from tracksegnet.utils import check_parms, get_color_list
from tracksegnet.simulate_tracks import run_track_simulation
from tracksegnet.experimental_tracks import extract_all_tracks, predict_states
from tracksegnet.generate_lstm_model import generate_lstm_model
from tracksegnet.analysis import compute_ptm, make_tracklet_lists, compute_all_msd,\
    plot_scatter_alpha_diffusion, plot_proportion, plot_displ, plot_angles, plot_vac

if __name__ == "__main__":

    START_TIME = datetime.now()

    PARMS_FILENAME = sys.argv[-1]
    # PARMS_FILENAME = 'parms.csv'
    PARMS_DF = pd.read_csv(PARMS_FILENAME, sep='\t').set_index('parms').squeeze().to_dict()
    PARMS_DF = check_parms(PARMS_DF)
    NSTATES = PARMS_DF['num_states']
    PARMS = {
        ## Path to the trajectories to analyze:
        'data_path': Path(PARMS_DF['data_path']),
        'track_format': 'MDF',

        ## Microscope settings:
        'time_frame': PARMS_DF['time_frame'], # Time between two frames in second
        'pixel_size': PARMS_DF['pixel_size'], # in micron

        ## Diffusive states:
        'num_states': NSTATES,
        # List of diffusive state:
        # motion parameters describing the behaviour of tracjectories within a state:
        # (1) the anomalous exponent alpha (expressing the confinement)
        # (2) the sigma value (related to the diffusion: sigma = np.sqrt((diffusion/unit)*2))
        'all_states': [{
            'diff': PARMS_DF[f'state_{state}_diff'],
            'alpha': PARMS_DF[f'state_{state}_alpha']
        } for state in range(1, NSTATES+1)],

        ## Restrictions on the track length:
        'length_threshold': 14,

        ## Trajectory simulation:
        'track_length_fixed': True,
        'track_length': 27,
        'num_simulate_tracks': 10000,
        # transition probabilities
        'ptm': np.array([[PARMS_DF[f'pt_{state1}_{state2}'] for state1 in range(1, NSTATES+1)]\
                        for state2 in range(1, NSTATES+1)]),
        'beta':100, # Mean of exponential distribution for length tracks

        ## Parameters to generate the LSTM model:
        'num_features': 6,
        'hidden_units': 200,
        'epochs': 1000, # Maximum number of epochs
        'patience': 15, # Parameter for the EarlyStopping function
        'batch_size': 2**5,
        'percent': {'val': 1/3, 'test':1/4},

        # Cell folders containing tracks.simple.mdf file to analyze
        'folder_names': [Path(f"{PARMS_DF['data_path']}/{folder}")\
                         for folder in os.listdir(PARMS_DF['data_path'])],

        # Colors and labels for plotting:
        'colors': get_color_list(NSTATES),

        # Figure saving:
        'fig_format': 'png',
        'fig_dpi': 200,
        'fig_transparent': False,
    }

    PARMS['unit_diff'] = (PARMS['pixel_size']**2)/PARMS['time_frame'] # in um**2/2
    STATES = f"{PARMS['num_states']}states=[" + '_'.join([f"({state['alpha']},{state['diff']})"\
        for _, state in enumerate(PARMS['all_states'])]) + "]"

    # Paths:
    PARMS.update({
        'simulated_track_path': Path(f"results/lstm_models/{PARMS['num_simulate_tracks']}_simulated_tracks"),
        'result_path': Path(f"results/{PARMS['data_path'].name}_{int(PARMS['time_frame']*1000)}ms_{STATES}"),
    })
    PARMS.update({
        'model_path': PARMS['simulated_track_path']/f'model_{STATES}',
    })
    os.makedirs(PARMS['model_path'], exist_ok=True)
    os.makedirs(PARMS['result_path'], exist_ok=True)

    ## BUILD LSTM MODEL
    ###################
    SIM_CSV = PARMS['simulated_track_path']/'simulated_tracks_df.csv'
    if not os.path.isfile(SIM_CSV):
        SIM_DF = run_track_simulation(PARMS)
    else:
        print('\nLoad simulated trajectories...')
        SIM_DF = pd.read_csv(SIM_CSV)

    if not os.path.isfile(PARMS['model_path']/'best_model.h5'):
        generate_lstm_model(SIM_DF, PARMS)

    ## PREDICT TRACKLET STATE
    #########################
    print('\nLoad trained model...')
    MODEL = load_model(PARMS['model_path']/'best_model.h5', compile=False)

    TRACK_CSV = PARMS['result_path']/f"{PARMS['data_path'].name}_tracks_df.csv"
    if not os.path.isfile(TRACK_CSV):
        TRACK_DF = extract_all_tracks(PARMS)
    else:
        print('\nLoad experimental trajectories...')
        TRACK_DF = pd.read_csv(TRACK_CSV)

    if 'state' not in TRACK_DF:
        predict_states(TRACK_DF, MODEL, PARMS)

    ## ANALYSIS
    ###########
    PTM_CSV = PARMS['result_path']/'transition_probabilities.csv'
    if not os.path.isfile(PTM_CSV):
        PTM = compute_ptm(TRACK_DF, PARMS)
    else:
        PTM = pd.read_csv(PTM_CSV)
    print(PTM)

    if not os.path.isfile(PARMS['result_path']/"motion_parms_state_1.csv"):
        TRACKLET_LISTS = make_tracklet_lists(TRACK_DF, PARMS)
        MOTION_PARMS = compute_all_msd(TRACKLET_LISTS, PARMS)
    else:
        TRACKLET_LISTS = make_tracklet_lists(TRACK_DF, PARMS)
        MOTION_PARMS = [pd.read_csv(PARMS['result_path']/f"motion_parms_state_{state+1}.csv") for state in range(PARMS['num_states'])]

    plot_proportion(TRACKLET_LISTS, PARMS)
    plot_scatter_alpha_diffusion(MOTION_PARMS, PARMS)
    plot_displ(TRACKLET_LISTS, PARMS)
    plot_angles(TRACKLET_LISTS, PARMS)
    plot_vac(TRACKLET_LISTS, PARMS)



"""Generate LSTM model

This module implements functions to generate a LSTM model for trajectory classification.
"""

# Third-party modules
import os
from datetime import datetime
from contextlib import redirect_stdout
import numpy as np
import matplotlib.pyplot as plt
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.models import Sequential, load_model
from tensorflow.keras.layers import LSTM, Dense, TimeDistributed, Bidirectional


def generate_lstm_model(sim_df, parms):
    """Prepares the dataset, defines the neural network, trains and evaluates the model.
    Best model is automatically saved.

    :param sim_df: Dataframe containing the simulated tracks for the training.
    :type sim_df: pd.DataFrame
    :param parms: Stored parameters containing global variables and instructions.
    :type parms: dict
    """
    print('\nTrain LSTM model...')
    start_time = datetime.now()

    ## Prepare the datasets
    feature_names = ['displ_x', 'displ_y', 'dist', 'mean_dist_1', 'mean_dist_2', 'angle']
    all_features = np.array([sim_df[sim_df['track_id'] == N][feature_names].to_numpy()[1:-1]\
        for N in sim_df['track_id'].unique()])
    all_true_states = np.array([sim_df[sim_df['track_id'] == N]['state'].to_numpy()[1:-1]\
        for N in sim_df['track_id'].unique()])
    # Split data into train and test set (with shuffeling)

    all_true_states = to_categorical(all_true_states, num_classes=parms['num_states'])

    # split in training and test data
    num_train = round((1-parms['percent']['test'])*len(all_features))
    train_set = {'features': all_features[:num_train], 'states': all_true_states[:num_train]}
    test_set = {'features': all_features[num_train:], 'states': all_true_states[num_train:]}


    ## Define the LSTM neural network
    model = Sequential()
    model.add(Bidirectional(LSTM(parms['hidden_units'], return_sequences=True),
                            input_shape=(None, parms['num_features']), merge_mode='concat'))
    model.add(TimeDistributed(Dense((parms['num_states']), activation='softmax'))) # Softmax to get
    # 'probability distribution'
    model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
    # Training stops once the model performance does not improve on a hold out validation dataset
    # The best mode lis saved
    callbacks = [EarlyStopping(monitor='val_loss', patience=parms['patience']),\
                 ModelCheckpoint(filepath=str(parms['model_path']/'best_model.h5'),
                                 monitor='val_loss', save_best_only=True)]
    history = model.fit(train_set['features'], train_set['states'], callbacks=callbacks,
                        epochs=parms['epochs'], batch_size=parms['batch_size'],
                        validation_split=parms['percent']['val'], shuffle=True)
    plot_training_curves(parms['model_path'], history, parms['patience'])
    with open(parms['model_path']/'model_summary.txt', 'w', encoding='utf-8') as file:
        with redirect_stdout(file):
            model.summary()

    ## evaluate the model
    # Predicting on test set and convert binary matrix into vector
    best_model = load_model(parms['model_path']/'best_model.h5')
    train_evaluation = best_model.evaluate(train_set['features'], train_set['states'], verbose=0)
    test_evaluation = best_model.evaluate(test_set['features'], test_set['states'], verbose=0)

    with open(parms['model_path']/'evaluation_model.txt', 'w', encoding='utf-8') as file:
        print(f'Train loss:\t{train_evaluation[0]:.2f}\tTrain accuracy:\t{train_evaluation[1]:.2f}',
              file=file)
        print(f'Test loss:\t{test_evaluation[0]:.2f}\tTest accuracy:\t{test_evaluation[1]:.2f}',
              file=file)
    predicted_test = best_model.predict(test_set['features'])
    predicted_states_test = predicted_test.argmax(axis=2) # Predicted transition states
    true_states_test = test_set['states'].argmax(axis=2) # Real transition states
    booleans = np.array(predicted_states_test == true_states_test, dtype=int)
    plot_accuracy_over_window(parms['model_path'], booleans, parms['track_length']-2)


    ## total running time
    time = datetime.now() - start_time
    hours = time.seconds // 3600
    minutes = time.seconds // 60 % 60
    print(f'\nTotal runtime: {hours}h{minutes:0>2d}')


def plot_training_curves(path, history, patience):
    """Plots loss and accuracy training curves.

    :param path: Output path.
    :type path: str
    :param history: History of the training to be plotted.
    :type history: History keras object
    :param patience: Value of the EarlyStopping criterion used here to trim the axis.
    :type patience: int
    """
    path = path/'training_plots'
    os.makedirs(path, exist_ok=True)

    x_axis = range(len(history.history['loss']))
    # Plot loss curve
    plt.plot(x_axis, history.history['loss'], label='Training')
    plt.plot(x_axis, history.history['val_loss'], label='Validation')
    plt.axvline(x_axis[-1]-patience, color='k', linestyle='--', label='best model')
    plt.title('Training and validation loss')
    plt.xlabel('Epochs')
    plt.ylabel('Loss')
    plt.legend(loc='best')
    plt.savefig(path/'loss_curve.png', bbox_inches='tight')
    plt.close()
    # Plot accuracy curve
    plt.plot(x_axis, history.history['accuracy'], label='Training')
    plt.plot(x_axis, history.history['val_accuracy'], label='Validation')
    plt.axvline(x_axis[-1]-patience, color='k', linestyle='--', label='best model')
    plt.title('Training and validation accuracy')
    plt.xlabel('Epochs')
    plt.ylabel('Accuracy')
    plt.legend(loc='best')
    plt.savefig(path/'accuracy_curve.png', bbox_inches='tight')
    plt.close()

def plot_accuracy_over_window(path, booleans, window_size):
    """Plots a bar graph of the accuracy gradient over the tracks.

    :param path: Output path.
    :type path: str
    :param booleans: Array of true and false based on 'predicted_states_test == true_states_test'.
    :type booleans: np.array
    :param window_size: Used window size.
    :type window_size: int
    """
    path = path/'training_plots'
    os.makedirs(path, exist_ok=True)
    # ws in parms['window_size'] and hu in parms['hidden_units']
    booleans_mean = booleans.mean(axis=0)
    plt.bar(np.arange(window_size)+1, booleans_mean)
    plt.gca().set_xticks(np.arange(window_size)+1)
    plt.xlabel('Window position')
    plt.ylabel('Accuracy')
    plt.title('Accuracy gradient over window (test set)', y=1.04)
    plt.gca().set_ylim([min(booleans_mean - 0.02), max(booleans_mean + 0.02)])
    plt.tight_layout()
    plt.savefig(path/f'accuracy_over_gradient_w{window_size}.png', bbox_inches='tight')
    plt.close()
